
// blossom.hpp - methods for interacting with blossom4 code
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#ifndef BLOSSOM_HPP
#define BLOSSOM_HPP

#include "graph.hpp"

#define GET_SLACK
#define TIMER_DEFINED
double get_time() { return 0.0; }
#include "PerfectMatching.h"
#include "PMimplementation.h"

// compute the slack of an edge
PerfectMatching::REAL
    PerfectMatching::GetSlack(const Edge* e) const
{
    REAL slack = e->slack;
    Node* u = e->head[0];
    Node* v = e->head[1];

    // outer '-' blossom: y and slack are swapped
    if (u->is_outer && u->is_blossom && u->flag == 1 &&
        ARC_TO_EDGE_PTR(u->match) == e) return 0;
    if (v->is_outer && v->is_blossom && v->flag == 1 &&
        ARC_TO_EDGE_PTR(v->match) == e) return 0;

    // find outer ancestors
    std::vector<Node*> u_stack, v_stack;
    while (!u->is_outer)
    {
        u_stack.push_back(u);
        u = u->blossom_parent;
    }
    while (!v->is_outer)
    {
        v_stack.push_back(v);
        v = v->blossom_parent;
    }
 
    if (u == v)  // edge internal to a blossom
    {
        // find children of LCA(u,v)
        while (u == v)
        {
            assert(!u_stack.empty());
            assert(!v_stack.empty());
            u = u_stack.back();
            v = v_stack.back();
            u_stack.pop_back();
            v_stack.pop_back();
        }

        if (!PriorityQueue<REAL>::isReset((Edge*)e))
        {
            assert(u->blossom_eps == v->blossom_eps);
            slack -= 2*u->blossom_eps;
        }

        return slack;
    }

    assert(u->is_outer && v->is_outer);
    if (u->flag == 0) slack -= u->tree->eps;
    else if (u->flag == 1) slack += u->tree->eps;
    if (v->flag == 0) slack -= v->tree->eps;
    else if (v->flag == 1) slack += v->tree->eps;
    return slack;
}

PerfectMatching::REAL
    PerfectMatching::GetSlack(const size_t idx) const
{
    assert(int(idx) < edge_num);
    return GetSlack(edges + idx);
}

// compute signed slack, taking neighbours into account
PerfectMatching::REAL
  PerfectMatching::ExtendedSlack(const size_t idx) const
{
  REAL slack = GetSlack(edges + idx);
  if (slack > PM_THRESHOLD) return slack;

  Edge* a;
  int dir;
  EdgeIterator I;

  Node* n = edges[idx].head[0];
  REAL min_slack = PM_INFTY;
  FOR_ALL_EDGES(n, a, dir, I)
    if (a != edges + idx && GetSlack(a) < min_slack)
      min_slack = a->slack;
  slack = min_slack;

  n = edges[idx].head[1];
  min_slack = PM_INFTY;
  FOR_ALL_EDGES(n, a, dir, I)
    if (a != edges + idx && GetSlack(a) < min_slack)
      min_slack = a->slack;
  slack += min_slack;

  return slack;
}

namespace Isinf
{
    class BlossomGraph : public Graph
    {
        typedef PerfectMatching::REAL REAL;
    public:
        PerfectMatching* pm;
    private:
        vector<REAL> edgelen;
        vector<int> edgelist;
        vector<bool> bicon;
        REAL scale;

        // writes out edge list of the
        // expanded dual as a vector
        void blossom_dual(void)
        {   
            undone();
            edgelist.clear();
            for (size_t n = 0; n < 2*nedges; ++n)
                edgelist.push_back(n);
            
            for (size_t n = 0; n < nodes.size(); ++n)
            {
                if (Edgelet* e = nodes[n].edges) do
                {
                    if (!e->done)  // traverse face
                    {
                        vector<int> idx;
                        Edgelet* f = e;
                        do
                        {
                            int i = 2*f->index;
                            if (f->other->done) ++i;
                            idx.push_back(i);
                            f->done = true;
                            f = f->other->next;
                        }
                        while (f != e);

                        for (size_t i = 0; i < idx.size() - 1; ++i)
                            for (size_t j = i + 1; j < idx.size(); ++j)
                            {
                                if (idx[i]/2 == idx[j]/2)
                                    // watch out: edge not biconnected
                                    bicon[idx[i]/2] = false;
                                else
                                {
                                    edgelist.push_back(idx[i]);
                                    edgelist.push_back(idx[j]);
                                }
                            }
                    }
                    e = e->next;
                }
                while (e != nodes[n].edges);
            }

            undone();
            return;
        }

        /*** DEPRECATED
        // specialisation for triangulated graphs
        void blossom_dual3(void)
        {   
            undone();
            edgelist.clear();
            for (size_t n = 0; n < 2*nedges; ++n)
                edgelist.push_back(n);
            edgelist.resize(6*nedges, 0);

            for (size_t n = 0; n < nodes.size(); ++n)
            {
                if (Edgelet* e = nodes[n].edges) do
                {
                    if (!e->done)  // traverse face
                    {
                        vector<int> idx;
                        Edgelet* f = e;
                        do
                        {
                            int i = 2*f->index;
                            if (f->other->done) ++i;
                            idx.push_back(i);
                            f->done = true;
                            f = f->other->next;
                        }
                        while (f != e);

                        // ensure it's triangular & biconnected
                        // BUGBUG is biconnectedness needed?
                        assert (idx.size() == 3);
                        assert(idx[0]/2 != idx[1]/2);
                        assert(idx[0]/2 != idx[2]/2);
                        assert(idx[1]/2 != idx[2]/2);

                        // use special indexing scheme
                        edgelist[2*(nedges + idx[0])] = idx[1];
                        edgelist[2*(nedges + idx[0]) + 1] = idx[2];
                        edgelist[2*(nedges + idx[1])] = idx[0];
                        edgelist[2*(nedges + idx[1]) + 1] = idx[2];
                        edgelist[2*(nedges + idx[2])] = idx[0];
                        edgelist[2*(nedges + idx[2]) + 1] = idx[1];
                    }
                    e = e->next;
                }
                while (e != nodes[n].edges);
            }

            undone();
            return;
        }
        ***/

    public:
        BlossomGraph(void) :
            Graph(), pm(0), bicon(nedges, true), scale(-1) {};
        BlossomGraph(istream& is) :
            Graph(is), pm(0), bicon(nedges, true), scale(-1) {};
        ~BlossomGraph() { if (pm) delete pm; }

        /*** DEPRECATED
        // compute signed slacks wrt. primal graph
        // NB graph must be triangulated!
        void extended_slacks(vector<REAL>& slacks)
        {
            slacks.clear();
            assert(2*edgelen.size() == edgelist.size());

            for (size_t i = 0; i < 3*nedges; ++i)
                slacks.push_back(pm->GetSlack(i));

            for (size_t i = nedges; i < 3*nedges; ++i)
                if (pm->GetSolution(i))
                    std::swap(slacks[nedges + edgelist[2*i]],
                        slacks[nedges + edgelist[2*i + 1]]);

            vector<REAL> tmp;
            for (size_t i = 0; i < nedges; ++i)
            {
                tmp.push_back(slacks[i]);
                size_t j = 2*(nedges + 2*i);
                tmp[i] += std::min(
                    slacks[nedges + edgelist[j]] + slacks[edgelist[j + 1]/2],
                    slacks[nedges + edgelist[j + 1]] + slacks[edgelist[j]/2]);
                // tmp[i] -= slacks[j/2];
                    // + slacks[edgelist[j]/2] + slacks[edgelist[j + 1]/2];
                j += 2;
                tmp[i] += std::min(
                    slacks[nedges + edgelist[j]] + slacks[edgelist[j + 1]/2],
                    slacks[nedges + edgelist[j + 1]] + slacks[edgelist[j]/2]);
                // tmp[i] -= slacks[j/2];
                    // + slacks[edgelist[j]/2] + slacks[edgelist[j + 1]/2];
            }
            slacks = tmp;
            // slacks.resize(nedges);
        }
        ***/

        // calculates ground state for disagreement costs e
        // (sets agreement on graph edges accordingly)
        void call_blossom(const double e[], const size_t n)
        {

            if (!pm)  // start from scratch
            {
                // scale edge costs
                edgelen.clear();

                if (REAL(1)/2 == 0)
                {
                    double psum = 0.0, nsum = 0.0;

                    for (size_t k = 0; k < n; ++k)
                    {
                        if (e[k] > 0.0)
                             psum += e[k];
                        else nsum -= e[k];
                    }
                    scale = -PM_INFTY/(1024*std::max(psum, nsum));
std::cerr << "scale factor: " << scale << std::endl;
                }

                for (size_t k = 0; k < n; ++k)
                    edgelen.push_back(REAL(scale*e[k]));

                // construct dual for blossom
                blossom_dual();
                while (edgelen.size() < edgelist.size()/2)
                    edgelen.push_back(0);

                pm = new PerfectMatching(2*nedges, edgelen.size());
                pm->options.dual_greedy_update_option = 0;
                pm->options.verbose = false;
                for (size_t k = 0; k < edgelen.size(); ++k)
                    pm->AddEdge(edgelist[2*k], edgelist[2*k + 1], edgelen[k]);
            }
            else  // incremental update
            {
                pm->StartUpdate();
                for (size_t k = 0; k < n; ++k)
                {
                    REAL delta = REAL(scale*e[k]);
                    delta -= edgelen[k];
                    if (delta)
                    {
                        edgelen[k] += delta;
                        pm->UpdateCost(k, delta);
                    }
                }
                pm->FinishUpdate();
            }

            pm->Solve();
            // read matching back, fixing
            // edges that are not biconnected
            for (size_t k = 0; k < nedges; ++k)
                edges[k]->agree = edges[k]->other->agree =
                    (bicon[k] ? pm->GetSolution(k) :
                        (k < n ? (e[k] >= 0.0) : true));
        }

        inline void call_blossom(const vector<double>& e)
        {
            call_blossom(&e[0], e.size());
        }

        // calculates marginal energies for each edge
        double energy_margins(vector<double>& e)
        {
            double nrg = 0.0;
            call_blossom(e);
            for (size_t k = 0; k < e.size(); ++k)
                if (!edges[k]->agree) nrg += e[k];

            const REAL delta = PM_INFTY/512;
            double n2[nedges];

            pm->StartUpdate();
            for (size_t i = 0; i < nedges; ++i)
            {
                pm->UpdateCost(i, (edges[i]->agree ? delta : -delta));
                pm->FinishUpdate();

                pm->Solve();
                // calculate energy (could be faster incrementally)
                double nrg2 = -nrg;
                for (size_t k = 0; k < e.size(); ++k)
                    if (!pm->GetSolution(k)) nrg2 += e[k];
                n2[i] = nrg2;

                pm->StartUpdate();
                pm->UpdateCost(i, (edges[i]->agree ? -delta : delta));
            }
            pm->FinishUpdate();

            size_t k;
            for (k = 0; k < e.size(); ++k) e[k] = n2[k];
            for (; k < nedges; ++k) e.push_back(n2[k]);
            return nrg;
        }

        // propagate node state along shortest-path tree rooted in source
        // returns shortest conflict loop through source if any, 0 otherwise
        // NB: changes state only on some of the nodes, no edges
        deque<size_t>& dijkstra_nodestate
            (const vector<double>& cost, int source = -1)
        {
            assert(cost.size() <= nedges);
            if (source == -1) source += nodes.size();
            deque<size_t>& loop = *new deque<size_t>;
            Graph::Edgelet* conflict = 0;
            undone();

            // build shortest-path tree from source
            PriorityVector<double> pv;
            double cdist = pv.end();
            pv[source] = 0.0;

            while (!pv.empty())
            {
                const size_t u = pv.top();

                if (Edgelet* e = nodes[u].edges) do
                {
                    const size_t v = e->other->node;
                    const double d = pv[u] + cost[e->index];

                    if (nodes[v].done)  // check for conflicts
                    {
                        if (nodes[v].state != (nodes[u].state == e->agree)
                            && d + pv[v] < cdist)
                        {
                            cdist = d + pv[v];
                            conflict = e;
                        }
                    }
                    else if (d < pv[v])  // relax the edge
                    {
                        pv[v] = d;
                        nodes[v].edges = e->other;  // marks shortest path
                        nodes[v].state = (nodes[u].state == e->agree);
                    }
                    e = e->next;
                }
                while (e != nodes[u].edges);

                pv[u].erase();
                nodes[u].done = true;
                if (conflict) break;
            }

            // find conflict loop
            if (conflict)
            {
                Edgelet* f;
                for (f = conflict; int(f->node) != source;
                    f = nodes[f->other->node].edges)
                    loop.push_back(f->index);
                for (f = nodes[conflict->node].edges; int(f->node) != source;
                    f = nodes[f->other->node].edges)
                    loop.push_front(f->index);

                while (loop.front() == loop.back())
                {
std::cerr << "dropping " << loop.front() << std::endl;
                    loop.pop_front();
                    loop.pop_back();
                }
            }

            /***
            // encode tree in edge "done" state (for edge quenching)
            for (size_t i = 0; i < edges.size(); ++i)
                edges[i]->done = edges[i]->other->done = true;
            for (size_t i = 0; i < nodes.size(); ++i)
                if (int(i) != source)
                    nodes[i].edges->done = nodes[i].edges->other->done = false;
            ***/

            undone();
            return loop;
        }
    };
}

#endif 

