
// bt.hpp - C++ wrapper for bt.f Fortran code
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#include "utils.hpp"

extern "C"
{
    void bt_(const int&, double[], double&, const double[],
        void(const int&, const double*, double&, double*),
        const double&, const double&, const int&, const int&, const int&,
        const int&, int&, const int[], double[], const int&);

}

extern void compfg(const int&, const double[], double&, double[]);

// x on input:  initial parameter values
// x on output: parameter values at minimum
// fm: lower bound on objective
// returns objective at minimum
double call_bt(std::vector<double>& x, const double fm = 0.0)
{
    // customizable parameters:
    const int iprint = 0;     // verbosity level
    const int maxit = 1000;   // max number of iterations
    const double eps = 1e-6;  // error tolerance
    int reset = 20;           // number of subgradients

    // internals - do not edit below here
    const int n = x.size();
    const int maxcom = (maxit*6)/5;  // max number of compfg calls

    int ifail = 0;            // error code
    double f, g[n];           // function value & gradient
    compfg(n, &x[0], f, g);   // to calculate intial f and g

    // space allocation
    int lwork;
    double* work = 0;
    reset *= 2;
    while (!work && reset >= 8)
    {
        reset /= 2;
        lwork = (reset + n + 9)*reset + 4*n + 10;
        work = new double[lwork];
    }
    int iwork[reset];

    if (!work)
    {
        std::cerr << Isinf::argv0 << "bt: out of memory\n";
        exit(42);
    }

    // call the optimizer
    bt_(n, &x[0], f, g, compfg, fm, eps, maxcom, maxit,
	      reset, iprint, ifail, iwork, work, lwork); 
    delete work;

    // report any errors
    if (ifail != 0) std::cerr << Isinf::argv0 << "bt: ";
    switch (ifail)
    {
        case 5:
            std::cerr << "QP solver failed" << std::endl;
            break;
            // exit(ifail);
        case 3:
            std::cerr << "T is too small (?)" << std::endl;
            break;
            // exit(ifail);
        case 2:
            std::cerr << "warning: maxit exceeded" << std::endl;
            break;
        case 1:
            std::cerr << "warning: maxcom exceeded" << std::endl;
            break;
        case 0:
            break;
        default:
            std::cerr << "unknown error " << ifail << std::endl;
            exit(ifail);
    }

    return f;
}

