
// fastPD.cpp - isinf interface to the FastPD code
// (c) 2009 Nic Schraudolph - all rights reserved

#include <stdio.h>
#include <fstream>
#include "utils.hpp"
#include "Fast_PD.h"

typedef CV_Fast_PD::Real Real;

int main(int argc, char *argv[])
{
    Isinf::argv0 = argv[0];
    if (argc != 3)
    {
        std::cerr << "usage: " << argv[0]
            << " <edgelist> <edgefeatures> < <params>" << std::endl;
        exit(1);
    }

    // read parameters from standard input
    std::vector<double> theta;
    std::cin >> theta;
    assert(!theta.empty());

    // read edge list from first argument
    std::ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    std::vector<int> edges;
    int num_nodes = 0;

    std::vector<int> x;
    arg1 >> x;
    while (x.size() >= 2)
    {
        if (x[0] == x[1])  // self-loop!
        {
            std::cerr << Isinf::argv0 << "error at node " << x[0]
                << ": self-loops not allowed" << std::endl;
            exit(3);
        }

        if (x[0] > x[1]) std::swap(x[0], x[1]);
        if (x[1] > num_nodes) num_nodes = x[1];
        edges.push_back(x[0]);
        edges.push_back(x[1]);

        x.clear();
        arg1 >> x;
    }

    if (edges.size() == 0)
    {
        std::cerr << Isinf::argv0 << "no valid edges found" << std::endl;
        exit(3);
    }

    // remove bias edges
    // BUGBUG assumes that bias node is last node,
    //        and that bias edges are listed last
    size_t num_edges;
    for (num_edges = edges.size() - 1; edges[num_edges] == num_nodes;
         num_edges -= 2);
    std::cerr << "read " << num_nodes + 1 << "-1 nodes, " <<
        edges.size()/2 << '-' << (edges.size() - num_edges)/2 << " edges\n";
    ++num_edges;
    num_edges /= 2;
    // only for full bias:
    // assert(int(edges.size() - 2*num_edges) == 2*num_nodes);

    // read features from second argument
    std::ifstream arg2(argv[2]);
    if (arg2.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[2] << "'\n";
        exit(4);
    }

    while (arg2.good())
    {
        std::vector<Real> wcosts;
        while (wcosts.size() < edges.size()/2)
        {
            std::vector<double> f;

            arg2 >> f;
            if (wcosts.empty() && f.empty()) exit(0);

            if (f.size() != theta.size())
            {
                std::cerr << Isinf::argv0 << "read " << f.size()
                    << " instead of " << theta.size()
                    << " features for edge " << wcosts.size() << std::endl;
                exit(5);
            }
            wcosts.push_back(f*theta);
        }

        // set up label costs
        Real label_dist[4] = {0, 1, 1, 0};
        std::vector<Real> lcosts(num_nodes, 0);
        for (int i = 0; i < num_nodes; ++i)
            lcosts.push_back(wcosts[num_edges + i]);

        // run the algorithm
        stdout = stderr;
        const int max_iters = 1000;
        CV_Fast_PD pd(num_nodes, 2, &lcosts[0], num_edges,
            &edges[0], label_dist, max_iters, &wcosts[0]);
        Isinf::CPUtimer *t = new Isinf::CPUtimer();
        pd.run();
        delete t;

        // compute energy
        size_t i;
        double nrg = 0.0;
        for (i = 0; i < num_edges; ++i)
            if (pd._pinfo[edges[2*i]].label != pd._pinfo[edges[2*i + 1]].label)
                nrg += wcosts[i];
        for (; i < edges.size()/2; ++i)
            if (pd._pinfo[edges[2*i]].label) nrg += wcosts[i];

        // output labels and energy
        for (int i = 0; i < num_nodes; ++i)
            std::cout << (pd._pinfo[i].label ? '+' : '-');
        std::cout << "- " << nrg << std::endl;
    }
}

