
// mima.cpp - compute edge min-marginals via Blossom
// (c) 2009 Nic Schraudolph - see COPYRIGHT.txt for details
//
// As suggested by: Dhruv Batra, Tsuhan Chen. Dynamic Planar-Cuts:
// Efficient Computation of Min-Marginals for Outer-Planar Models.
// Workshop on Discrete Optimization in Machine Learning, NIPS 2009. 
// http://www.ece.cmu.edu/~dbatra/publications/assets/batra_discml_nips09.pdf

#include <iomanip>
#include <fstream>
#include "blossom.hpp"

int main(int argc, char** argv)
{
    if (argc != 3)
    {
        std::cerr << "usage: " << argv[0] << " <embedding> "
            "<edgefeatures> < <params> > <min-marginals>" << std::endl;
        exit(1);
    }
    Isinf::argv0 = argv[0];
    
    // read parameters from standard input
    std::vector<double> theta;
    std::cin >> theta;
    assert(!theta.empty());

    // read graph from first argument
    std::ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    Isinf::BlossomGraph g(arg1);
    assert(g.is_connected());
    if (g.genus() > 0)
        std::cerr << Isinf::argv0 << "warning: graph has genus "
            << g.genus() << std::endl;
    size_t n = g.angulate();

    // read features from second argument
    std::ifstream arg2(argv[2]);
    if (arg2.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[2] << "'\n";
        exit(3);
    }

    while (arg2.good())
    {
        std::vector<double> e;

        while (e.size() < g.nedges - n)
        {
            std::vector<double> f;

            arg2 >> f;
            if (e.empty() && f.empty()) exit(0);

            if (f.size() != theta.size())
            {
                std::cerr << Isinf::argv0 << "read " << f.size()
                    << " instead of " << theta.size()
                    << " features for edge " << e.size() << std::endl;
                exit(4);
            }
            e.push_back(f*theta);
        }

        // find & print marginal energies
        Isinf::CPUtimer *t = new Isinf::CPUtimer;
        double nrg = g.energy_margins(e);
        delete t;
        std::cerr << nrg << std::endl;
        std::cout << e << std::endl;
    }
}

