
// mle.cpp - perform maximum likelihood CRF parameter estimation 
// (c) 2008, 2009 Nic Schraudolph - see COPYRIGHT.txt for details 

#include <fstream>
#include <sstream>
#include "lbfgs.hpp"
#include "superlu.hpp"

typedef struct
{
    std::vector<bool> labels;
    std::vector<std::vector<double> > features;
} Pattern;

std::vector<Pattern> train;

Isinf::Graph *graph;
Isinf::SuperLU *lu;

double lambda = 1.0;

extern double call_lbfgs(std::vector<double>&);


// given parameters x, compute gradient g and return function value
double compfg(const std::vector<double>& x, std::vector<double>& g)
{
    size_t i, j, k;
    std::vector<double> e;
    assert(g.size() == x.size());

    // regularization term
    double f = x*x;
    f *= train.size()*lambda/2.0;
    for (i = 0; i < g.size(); ++i)
        g[i] = train.size()*lambda*x[i];

    for (i = 0; i < train.size(); ++i)
    {
        // compute edge energies
        e.clear();
        for (j = 0; j < train[i].features.size(); ++j)
        {
            double sum = x*train[i].features[j];
            e.push_back(exp(sum));

            if (train[i].labels[graph->edges[j]->node] !=
                train[i].labels[graph->edges[j]->other->node])
            {
                for (k = 0; k < g.size(); ++k)
                    g[k] += train[i].features[j][k];
            }
            else f -= sum;
        }

        f += lu->marginals(e)/2.0;
        for (j = 0; j < train[i].features.size(); ++j)
            for (k = 0; k < g.size(); ++k)
                g[k] -= e[j]*train[i].features[j][k];
    }
    return f;
}

int main(int argc, char** argv)
{
    if (argc < 3 || argc > 4)
    {
        std::cerr << "usage: " << argv[0] << " <embedding> <edgefeatures>"
            " [<reg>] < <nodelabels>  > <params>" << std::endl;
        exit(1);
    }
    Isinf::argv0 = argv[0];

    // read graph from first argument
    std::ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    Isinf::Graph g(arg1);
    graph = &g;
    assert(graph->is_connected());
    assert(graph->genus() == 0);

    // triangulate & orient graph
    size_t n = graph->angulate(3);
    bool odd = graph->make_odd();
    assert(odd);

    // create SuperLU interface
    Isinf::SuperLU slu(g);
    lu = &slu;

    // read optional regularization parameter
    if (argc == 4)
    {
        std::istringstream arg3(argv[3]);
        arg3 >> lambda;
        if (arg3.fail() || lambda < 0.0)
        {
            std::cerr << "invalid regularization parameter"
                " (third argument)" << std::endl;
            exit(3);
        }
    }

    // read features from second argument
    std::ifstream arg2(argv[2]);
    if (arg2.fail())
    {
        std::cerr << Isinf::argv0 << "cannot open file '" << argv[2] << "'\n";
        exit(4);
    }

    std::vector<double> f;
    std::vector<std::vector<double> > e;
    Pattern p;
    size_t d = 0;

    while (arg2.good())
    {
        while (e.size() < graph->nedges - n)
        {
            arg2 >> f;
            if (e.empty() && f.empty()) break;

            if (d == 0) d = f.size();
            else if (f.size() != d)
            {
                std::cerr << Isinf::argv0 << "read " << f.size()
                    << " instead of " << d
                    << " features for edge " << e.size() << std::endl;
                exit(3);
            }
            e.push_back(f);
            f.clear();
        }
        if (e.empty()) break;
        p.features = e;
        e.clear();

        // read node labels from standard inpout
        size_t i;
        std::string s;
        getline(std::cin, s);
        assert(!std::cin.fail());
        assert(s.size() >= graph->nodes.size());
        for (i = 0; i < graph->nodes.size(); ++i)
        {
            assert(s[i] == '+' || s[i] == '-');
            p.labels.push_back(s[i] == '+');
        }
        if (i < s.size())
            assert(s[i] != '+' && s[i] != '-');

        train.push_back(p);
        p.labels.clear();
    }

    // perform the optimisation
    std::vector<double> x(d, 0.0);
    Isinf::CPUtimer *t = new Isinf::CPUtimer;
    call_lbfgs(x);
    delete t;
    std::cout << x << std::endl;
}

