
// mplp.cpp - isinf interface to the MPLP code
// (c) 2010 Nic Schraudolph - all rights reserved

#include <iomanip>
#include "utils.hpp"

// these globals are needed for tighten_main.cpp

int niter = 1000;
int niter_later = 100;
int nclus_to_add = 5;
double obj_del_thr = 1e-4;
double int_gap_thr = 1e-5;

int graph_type;
int numRows = 0;
int numCols = 0;

clock_t start, finish;
double mplp_time;

// hack alert: chopping tighten_main.cpp's
// main() into two ordinary functions...
#define main(ARGC, ARGV) not_used(ARGC, ARGV)
#define mplp(A,B,C,D,E) *x; return 0;} int run(MPLPAlg &mplp) {
#define time mplp_time  // conflicts with time.h
#include "tighten_main.cpp"

// my interface class to MPLP
class MPLP : public MPLPAlg
{
public:
    // efficient constructor
    MPLP(const size_t num_nodes, const size_t num_edges,
         const vector<int>& edges, const vector<double>& cost) :
         MPLPAlg("/dev/null", "/dev/null", "/dev/null", "/dev/null", "/dev/null")
    {
        assert(cost.size() == num_edges + num_nodes);

        for (size_t i = 0; i < num_edges; ++i)
        {
            vector<int> tmp;
            tmp.push_back(edges[2*i]);
            tmp.push_back(edges[2*i + 1]);

            m_all_region_inds.push_back(tmp);
            m_all_intersects.push_back(tmp);
            m_intersect_map.insert(pair<pair<int,int>,int>(
                pair<int,int>(tmp[0], tmp[1]), m_all_intersects.size() - 1));

            tmp.insert(tmp.begin(), i);
            tmp[1] += num_edges;
            tmp[2] += num_edges;
            m_all_region_intersects.push_back(tmp);
        }

        for (size_t i = 0; i < num_nodes; ++i)
        {
            vector<int> tmp;
            tmp.push_back(i);
            m_all_region_inds.push_back(tmp);
            m_all_intersects.push_back(tmp);
            tmp.front() += num_edges;
            m_all_region_intersects.push_back(tmp);
        }

        // NB: assumes all messages are initialized to zero 
        for (size_t i = 0; i < m_all_intersects.size(); ++i)
        {
            vector<int> tmp(m_all_intersects[i].size(), 2);
            MulDimArr mda(tmp);
            mda = 0;
            m_sum_into_intersects.push_back(mda);
        }

        m_var_sizes.resize(num_nodes, 2);
        m_decoded_res.resize(num_nodes, 0);

        // initialise regions
        vector<int> tmp(2, 2);
        MulDimArr lambda(tmp);
        for (size_t i = 0; i < num_edges; ++i)
        {
            lambda[0] = lambda[3] = cost[i];
            lambda[1] = lambda[2] = 0.0;
            Region r(m_all_region_inds[i], lambda, m_all_intersects,
                m_all_region_intersects[i], m_var_sizes);
            m_all_regions.push_back(r);
        }
        tmp.resize(1);
        MulDimArr lambda2(tmp);
        for (size_t i = num_edges; i < num_edges + num_nodes; ++i)
        {
            lambda2[0] = cost[i];
            lambda2[1] = 0.0;
            Region r(m_all_region_inds[i], lambda2, m_all_intersects,
                m_all_region_intersects[i], m_var_sizes);
            m_all_regions.push_back(r);
        }
    }
};

/***
        vector<int> m_decoded_res;
        vector<double> m_objhist;
        vector<double> m_inthist;
        vector<double> m_timehist;
***/

#undef main

int main(int argc, char *argv[])
{
    Isinf::argv0 = argv[0];
    if (argc != 3 && argc != 5)
    {
        cerr << "usage: " << argv[0] << " <edgelist> <edgefeatures>"
             "[<rows> <cols>] < <params>" << endl;
        exit(1);
    }
    if (argc == 5)
    {
        graph_type = GRAPH_TYPE_GRID;
        numRows = atoi(argv[3]);
        numCols = atoi(argv[4]);
        if (numRows <= 0 || numCols <= 0)
        {
            cerr << "row/column number must be positive" << endl;
            exit(2);
        }
    }
    else graph_type = GRAPH_TYPE_DEFAULT;

    // read parameters from standard input
    vector<double> theta;
    cin >> theta;
    assert(!theta.empty());

    // read edge list from first argument
    ifstream arg1(argv[1]);
    if (arg1.fail())
    {
        cerr << Isinf::argv0 << "cannot open file '" << argv[1] << "'\n";
        exit(2);
    }

    vector<int> edges;
    int num_nodes = 0;

    vector<int> x;
    arg1 >> x;
    while (x.size() >= 2)
    {
        if (x[0] == x[1])  // self-loop!
        {
            cerr << Isinf::argv0 << "error at node " << x[0]
                << ": self-loops not allowed" << endl;
            exit(3);
        }

        if (x[0] > x[1]) swap(x[0], x[1]);
        if (x[1] > num_nodes) num_nodes = x[1];
        edges.push_back(x[0]);
        edges.push_back(x[1]);

        x.clear();
        arg1 >> x;
    }

    if (edges.size() == 0)
    {
        cerr << Isinf::argv0 << "no valid edges found" << endl;
        exit(3);
    }

    // remove bias edges
    // BUGBUG assumes that bias node is last node,
    //        and that bias edges are listed last
    size_t num_edges = edges.size() - 1;
    for (; edges[num_edges] == num_nodes; num_edges -= 2);
    cerr << "read " << num_nodes + 1 << "-1 nodes, " <<
        edges.size()/2 << '-' << (edges.size() - num_edges)/2 << " edges\n";
    ++num_edges;
    num_edges /= 2;
    // only for full bias:
    // assert(int(edges.size() - 2*num_edges) == 2*num_nodes);

    // read features from second argument
    ifstream arg2(argv[2]);
    if (arg2.fail())
    {
        cerr << Isinf::argv0 << "cannot open file '" << argv[2] << "'\n";
        exit(4);
    }

    while (arg2.good())
    {
        vector<double> costs;
        while (costs.size() < edges.size()/2)
        {
            vector<double> f;

            arg2 >> f;
            if (costs.empty() && f.empty()) exit(0);

            if (f.size() != theta.size())
            {
                cerr << Isinf::argv0 << "read " << f.size()
                    << " instead of " << theta.size()
                    << " features for edge " << costs.size() << endl;
                exit(5);
            }
            costs.push_back(f*theta);
        }

        ofstream of("/dev/null");  // redirect cout to /dev/null
        streambuf* cout_sb = cout.rdbuf(of.rdbuf());
        assert(costs.size() <= num_edges + num_nodes);
        costs.resize(num_edges + num_nodes, 0);
        // set up MPLP data structures
        MPLP my_mplp(num_nodes, num_edges, edges, costs);
        cout.rdbuf(cerr.rdbuf());  // redirect cout to cerr

        Isinf::CPUtimer *t = new Isinf::CPUtimer();
        run((MPLPAlg&) my_mplp);   // run the algorithm
        delete t;
        cout.rdbuf(cout_sb);       // restore cout to normal

        // compute energy
        size_t i;
        double nrg = 0.0;
        for (i = 0; i < num_edges; ++i)
            if (my_mplp.m_decoded_res[edges[2*i]] !=
                my_mplp.m_decoded_res[edges[2*i + 1]]) nrg += costs[i];
        for (; i < edges.size()/2; ++i)
            if (my_mplp.m_decoded_res[edges[2*i]]) nrg += costs[i];

        // output labels and energy
        for (int i = 0; i < num_nodes; ++i)
            cout << (my_mplp.m_decoded_res[i] ? '+' : '-');
        cout << "- " << setprecision(12) << nrg << endl;
    }
}

