
// superlu.hpp - interface class to SuperLU package
// (c) 2009 Nic Schraudolph - see COPYRIGHT.txt for details

#include <superLU/SRC/slu_ddefs.h>
#include "graph.hpp"

extern int dgstrs_incomplete_solve;

namespace Isinf
{
    class SuperLU
    {
    private:
        int info;
        size_t size;
        const size_t prefactor;
        SuperMatrix A, L, U;
        SuperLUStat_t stat;
        mem_usage_t mem_usage;
        superlu_options_t options;
        double rpg, rcond;
        char equed;

        std::vector<double> a, r, c;
        std::vector<int> arow, xa;
        std::vector<int> asup, asub;
        std::vector<double> scal_r, scal_c;
        std::vector<int> perm_r, perm_c, etree;
        std::vector<double> ferr, berr;

        void init(void)
        {
            a.reserve(6*size);
            arow.reserve(6*size);
            xa.reserve(2*size + 1);
            asup.reserve(size);
            asub.reserve(size);

            r.resize(2*size);
            c.resize(2*size);
            perm_r.resize(2*size);
            perm_c.resize(2*size);
            etree.resize(2*size);
            ferr.resize(size);
            berr.resize(size);

            StatInit(&stat);
            set_default_options(&options);
            options.ColPerm = COLAMD;
            options.PrintStat = NO;
            options.Fact = DOFACT;
            options.Equil = YES;

            info = -1;
        }

    public:
        SuperLU(Graph& g, size_t p = 0) : size(g.nedges), prefactor(p),
            scal_r(2*g.nedges, 1.0), scal_c(2*g.nedges, 1.0)
        {
            init();

            // make SuperMatrix from g
            size_t j, k;
            int n = 0;

            for (size_t i = 0; i < 2*size; ++i)
            {
                xa.push_back(n);

                Graph::Edgelet* e = g.edges[i/2];
                if (e->head) e = e->other;

                // look right
                j = 2*(e->prev->index);
                if (!e->prev->head) ++j;
                k = 2*(e->other->next->index);
                if (e->other->next->head) ++k;
                assert(j != k);

                a.push_back(1.0);
                arow.push_back(j);
                if (j == i + 1) asub.push_back(n);
                ++n;

                a.push_back(-1.0);
                arow.push_back(k);
                if (k == i + 1) asub.push_back(n);
                ++n;

                if (j != i + 1 && k != i + 1)
                {
                    a.push_back(0.0);
                    arow.push_back(i + 1);
                    asub.push_back(n);
                    ++n;
                }

                ++i;
                xa.push_back(n);

                // look left
                j = 2*(e->next->index);
                if (e->next->head) ++j;
                k = 2*(e->other->prev->index);
                if (!e->other->prev->head) ++k;
                assert(j != k);

                a.push_back(-1.0);
                arow.push_back(j);
                if (j == i - 1) asup.push_back(n);
                ++n;

                a.push_back(1.0);
                arow.push_back(k);
                if (k == i - 1) asup.push_back(n);
                ++n;

                if (j != i - 1 && k != i - 1)
                {
                    a.push_back(0.0);
                    arow.push_back(i - 1);
                    asup.push_back(n);
                    ++n;
                }
            }
            xa.push_back(n);

            assert(asup.size() == size);
            assert(asub.size() == size);
            dCreate_CompCol_Matrix(&A, 2*size, 2*size, n,
                &a[0], &arow[0], &xa[0], SLU_NC, SLU_D, SLU_GE);
        }

        ~SuperLU()
        {
            // Destroy_CompCol_Matrix(&A); 
            Destroy_CompCol_Matrix(&U); 
            Destroy_SuperNode_Matrix(&L); 
            StatFree(&stat);
        }

        void print_matrix(char* s, SuperMatrix* M)
        {
            assert(M->Dtype == SLU_D);
            bool Stype_is_NC_or_SC = false;

            if (M->Stype == SLU_NC)
            {
                NCformat *Mstore = (NCformat *) M->Store;
                double *Mval = (double*)(Mstore->nzval);
                for (int j = 0; j < M->ncol; ++j)
                {
                    for (int k = Mstore->colptr[j];
                        k < Mstore->colptr[j+1]; ++k)
                        std::cout << s << '(' << Mstore->rowind[k] << ','
                            << j << ") = " << Mval[k] << std::endl;
                }
            }
            else if (M->Stype == SLU_SC)
            {
                SCformat *Mstore = (SCformat *) M->Store;
                double *Mval = (double*)(Mstore->nzval);
                for (int j = 0; j < M->ncol; ++j)
                {
                    int l = Mstore->rowind_colptr[
                        Mstore->sup_to_col[Mstore->col_to_sup[j]]];
                    for (int k = Mstore->nzval_colptr[j];
                        k < Mstore->nzval_colptr[j+1]; ++k, ++l)
                        std::cout << s << '(' << Mstore->rowind[l] << ','
                            << j << ") = " << Mval[k] << std::endl;
                }
            }
            else assert(Stype_is_NC_or_SC);
        }

        // sanity check: LU = A ?
        void remultiply()
        {
            SCformat *Lstore = (SCformat *) L.Store;
            NCformat *Ustore = (NCformat *) U.Store;
            NCformat *Astore = (NCformat *) A.Store;
            double *Lval = (double*)(Lstore->nzval);
            double *Uval = (double*)(Ustore->nzval);
            double *Aval = (double*)(Astore->nzval);

            // compute col j of LU
            std::multiset<std::pair<int, double> > col;
            for (int j = 0; j < int(2*size); ++j)
            {
                col.clear();

                // scan col perm_c[j] of U
                for (int i = Ustore->colptr[perm_c[j]];
                    i < Ustore->colptr[perm_c[j]+1]; ++i)
                {
                    double f = Uval[i];
                    int r = Ustore->rowind[i];

                    // insert U_r,perm_c[j] times col r of L into col
                    col.insert(std::pair<int, double>(r, f));
                    int k = Lstore->rowind_colptr[
                        Lstore->sup_to_col[Lstore->col_to_sup[r]]];
                    for (int l = Lstore->nzval_colptr[r];
                        l < Lstore->nzval_colptr[r+1]; ++k, ++l)
                        if (Lstore->rowind[k] > r)
                            col.insert(std::pair<int, double>
                                (Lstore->rowind[k], f*Lval[l]));
                }

                // scan col perm_c[j] of L for non-subdiagonal
                // elements, which really belong to U (sigh)
                int m = Lstore->rowind_colptr[
                    Lstore->sup_to_col[Lstore->col_to_sup[perm_c[j]]]];
                for (int i = Lstore->nzval_colptr[perm_c[j]];
                    i < Lstore->nzval_colptr[perm_c[j]+1]; ++i, ++m)
                {
                    double f = Lval[i];
                    int r = Lstore->rowind[m];

                    if (r <= perm_c[j])
                    {
                        // insert U_r,perm_c[j] times col r of L into col
                        col.insert(std::pair<int, double>(r, f));
                        int k = Lstore->rowind_colptr[
                            Lstore->sup_to_col[Lstore->col_to_sup[r]]];
                        for (int l = Lstore->nzval_colptr[r];
                            l < Lstore->nzval_colptr[r+1]; ++k, ++l)
                            if (Lstore->rowind[k] > r)
                                col.insert(std::pair<int, double>
                                    (Lstore->rowind[k], f*Lval[l]));
                    }
                }

                // compare with col j of A
                int k = Astore->colptr[j];
                if (col.empty())
                {
                    assert(k == Astore->colptr[j+1]);
                    continue;
                }

                std::set<std::pair<int, double> > c2;
                for (; k < Astore->colptr[j+1]; ++k)
                    c2.insert(std::pair<int, double>(perm_r[Astore->rowind[k]], Aval[k]));


                std::multiset<std::pair<int, double> >::iterator i = col.begin();
                std::set<std::pair<int, double> >::iterator l = c2.begin();

#define TOL 1e-13

                int prev = i->first;
                double val = i->second;
                for (++i; i != col.end(); ++i)
                {
                    if (i->first == prev) val += i->second;
                    else
                    {
                        if (std::abs(val) > TOL)  // verify entry in A
                        {
                            while (l != c2.end() && l->first < prev)
                            {
                                std::cerr << "Only in A(" << l->first << ','
                                    << j << "): " << l->second << std::endl;
                                ++l;
                            }
                            if (l == c2.end() || l->first > prev)
                            {
                                std::cerr << "Only in LU(" << prev << ','
                                    << j << "): " << val << std::endl;
                            }
                            else  // same cell: compare values
                            {
                                assert(std::abs((l->second - val)/l->second) < TOL);
                                ++l;
                            }
                        }
                        prev = i->first;
                        val = i->second;
                    }
                }
                if (std::abs(val) > TOL)  // verify entry in A
                {
                    while (l != c2.end() && l->first < prev)
                    {
                        std::cerr << "Only in A(" << l->first << ','
                            << j << "): " << l->second << std::endl;
                        ++l;
                    }
                    if (l == c2.end() || l->first > prev)
                    {
                        std::cerr << "Only in LU(" << prev << ','
                            << j << "): " << val << std::endl;
                    }
                    else  // same cell: compare values
                    {
                        assert(std::abs((l->second - val)/l->second) < TOL);
                        ++l;
                    }

                    while (l != c2.end())
                    {
                        std::cerr << "Only in A(" << l->first << ','
                            << j << "): " << l->second << std::endl;
                        ++l;
                    }
                }
            }
        }

        // compute log determinant for potentials e
        double logdet(const std::vector<double>& e)
        {
            assert(e.size() <= size);
            size_t i;

            // add pots to super/subdiagonal
            for (i = 0; i < e.size(); ++i)
            {
                a[asub[i]] -= e[i]*scal_r[2*i+1]*scal_c[2*i];
                a[asup[i]] += e[i]*scal_r[2*i]*scal_c[2*i+1];
            }
            for (; i < size; ++i)
            {
                a[asub[i]] -= scal_r[2*i+1]*scal_c[2*i];
                a[asup[i]] += scal_r[2*i]*scal_c[2*i+1];
            }

            // perform LU factorisation
            SuperMatrix X;
            dCreate_Dense_Matrix(&X, 2*size, 0,
                NULL, 2*size, SLU_DN, SLU_D, SLU_GE);

            if (options.Fact != DOFACT)
                options.Fact = SamePattern;
            dgssvx(&options, &A, &perm_c[0], &perm_r[0], &etree[0], &equed,
                &r[0], &c[0], &L, &U, NULL, 0, &X, &X, &rpg,
                &rcond, &ferr[0], &berr[0], &mem_usage, &stat, &info);
            options.Fact = FACTORED;

            if (info)  // error
            {
                std::cerr << argv0 << "dgssv(): ";
                if (info <= int(2*size))
                    std::cerr << "matrix singular at row " << info;
                else
                    std::cerr << "could not allocate " << info - 2*size
                        << " bytes of memory";
                std::cerr << std::endl;
                exit(info);
            }

            // record equilibriation
            if (equed == 'R' || equed == 'B')
                for (i = 0; i < 2*e.size(); ++i) scal_r[i] *= r[i];
            if (equed == 'C' || equed == 'B')
                for (i = 0; i < 2*e.size(); ++i) scal_c[i] *= c[i];

            // restore super/subdiagonal
            for (i = 0; i < e.size(); ++i)
            {
                a[asub[i]] += e[i]*scal_r[2*i+1]*scal_c[2*i];
                a[asup[i]] -= e[i]*scal_r[2*i]*scal_c[2*i+1];
            }
            for (; i < size; ++i)
            {
                a[asub[i]] += scal_r[2*i+1]*scal_c[2*i];
                a[asup[i]] -= scal_r[2*i]*scal_c[2*i+1];
            }

            // compute log-determinant
            double lndet = 0.0;
            for (i = 0; i < 2*size; ++i)
            {
                lndet -= log(scal_r[i]);
                lndet -= log(scal_c[i]);
            }

            // diagonal of U is actually stored in L (sigh)
            SCformat *Lstore = (SCformat *) L.Store;
            double *Lval = (double*)(Lstore->nzval);

            for (int k = 0; k <= Lstore->nsuper; ++k)
            {
                int col = Lstore->sup_to_col[k];
                for (int j = col; j < Lstore->sup_to_col[k+1]; ++j)
                {
                    int d = Lstore->nzval_colptr[j];
                    for (int i = Lstore->rowind_colptr[col];
                        i < Lstore->rowind_colptr[col+1]; ++i, ++d)
                        if (Lstore->rowind[i] == j) 
                            lndet += log(std::abs(Lval[d]));
                }
            }
            return lndet;
        }

        // return i'th subdiagonal inverse element,
        // using previous factorisation in logdet()
        double inverse(size_t i)
        {
            assert(i < size);
            assert(options.Fact == FACTORED);

            // create rhs vector
            std::vector<double> rhs(2*size, 0.0);
            rhs[2*i + 1] = 1.0;
            SuperMatrix X;
            dCreate_Dense_Matrix(&X, 2*size, 1,
                &rhs[0], 2*size, SLU_DN, SLU_D, SLU_GE);

            // triangular solve
            dgstrs_incomplete_solve = 1;
            dgssvx(&options, &A, &perm_c[0], &perm_r[0], &etree[0], &equed,
                &scal_r[0], &scal_c[0], &L, &U, NULL, 0, &X, &X, &rpg,
                &rcond, &ferr[0], &berr[0], &mem_usage, &stat, &info);
            dgstrs_incomplete_solve = 0;

            if (info)  // error
            {
                std::cerr << argv0 << "dgssv(): ";
                if (info <= int(2*size))
                    std::cerr << "matrix singular at row " << info;
                else
                    std::cerr << "could not allocate " << info - 2*size
                        << " bytes of memory";
                std::cerr << std::endl;
                exit(info);
            }

            return rhs[2*i];
        }

        // given potentials e, compute all marginals
        // into e, and also return log-determinant
        double marginals(std::vector<double>& e)
        {
// CPUtimer *t = new CPUtimer("factoring");
            double lndet = logdet(e);
// delete t;
            assert(options.Fact == FACTORED);
            size_t i;

            // create rhs matrix stripe
            std::vector<double> rhs;
            size_t cols = 4*e.size();
            if (cols > 128) cols = 128;

            bool okay = false;
            while (!okay)
            {
                try
                {
                    okay = true;
                    rhs.resize(2*size*cols);
                }
                catch (bad_alloc&)
                {
                    if (cols == 1)
                    {
                        std::cerr << argv0 << "out of memory, giving up"
                            << std::endl;
                        exit(1);
                    }
                    else
                    {
                        cols /= 2;
                        std::cerr << argv0 << "retrying with "
                            << cols << " columns" << std::endl;
                        okay = false;
                    }
                }
            }
            cols /= 4;
            if (cols == 0) cols = 1;
            rhs.resize(2*size*cols);

            SuperMatrix X;
            dCreate_Dense_Matrix(&X, 2*size, cols,
                &rhs[0], 2*size, SLU_DN, SLU_D, SLU_GE);

            // stripe-wise solve for marginals
            dgstrs_incomplete_solve = 1;
            for (size_t base = 0; base < e.size(); base += cols)
            {
// std::cerr << '\r' << base << '/' << e.size() << "   ";

                // last stripe may not be full size
                if (base + cols > e.size())
                {
                    cols = e.size() - base;
                    dCreate_Dense_Matrix(&X, 2*size, cols,
                        &rhs[0], 2*size, SLU_DN, SLU_D, SLU_GE);
                }

                // set up rhs to get stripe of inverse
                bzero(&rhs[0], rhs.size()*sizeof(double));
                for (i = 2*base + 1; i < rhs.size(); i += 2*(size + 1))
                    rhs[i] = 1.0;

                // triangular solve
                dgssvx(&options, &A, &perm_c[0], &perm_r[0], &etree[0], &equed,
                    &scal_r[0], &scal_c[0], &L, &U, NULL, 0, &X, &X, &rpg,
                    &rcond, &ferr[0], &berr[0], &mem_usage, &stat, &info);

                if (info)  // error
                {
                    std::cerr << argv0 << "dgssv(): ";
                    if (info <= int(2*size))
                        std::cerr << "matrix singular at row " << info;
                    else
                        std::cerr << "could not allocate " << info - 2*size
                            << " bytes of memory";
                    std::cerr << std::endl;
                    exit(info);
                }

                // calculate marginals
                for (i = 0; i < cols; ++i)
                {
                    e[base + i] *= rhs[2*(base + i*(size + 1))];
                    ++(e[base + i]);
                }
            }

            dgstrs_incomplete_solve = 0;
            return lndet;
        }

        void printStats(void)
        {
            if (info == -1)
            {
                std::cerr << argv0 << "solve() never called\n";
                exit(-1);
            }
            StatPrint(&stat);
            if (info <= int(2*size))  // factorisation complete
            {
                SCformat *Lstore = (SCformat *) L.Store;
                NCformat *Ustore = (NCformat *) U.Store;
                std::cerr << "nnz: L " << Lstore->nnz
                    << ", U " << Ustore->nnz << ", total "
                    << Lstore->nnz + Ustore->nnz - 2*size << std::endl;

                // dQuerySpace(&L, &U, &mem_usage);
                std::cerr << "memory: L\\U " << mem_usage.for_lu/1e6
                    << " MB, total " << mem_usage.total_needed/1e6
                    << " MB, " << mem_usage.expansions << " expansions\n";
            }
        }
    };
}

